"use strict";
const extension_composer_factory_1 = require("./extension-composer-factory");
const _ = require("lodash");
const fs = require("fs");
const glob = require("glob");
const jsonInPlace = require("json-in-place");
const loc = require("./loc");
const path = require("path");
const Q = require("q");
const qfs = require("../../../lib/qfs");
const trace = require("../../../lib/trace");
const version = require("../../../lib/version");
/**
 * Facilitates the gathering/reading of partial manifests and creating the merged
 * manifests (one for each manifest builder)
 */
class Merger {
    /**
     * constructor. Instantiates one of each manifest builder.
     */
    constructor(settings) {
        this.settings = settings;
        this.manifestBuilders = [];
    }
    gatherManifests() {
        trace.debug('merger.gatherManifests');
        if (this.settings.manifestGlobs && this.settings.manifestGlobs.length > 0) {
            const globs = this.settings.manifestGlobs.map(p => path.isAbsolute(p) ? p : path.join(this.settings.root, p));
            trace.debug('merger.gatherManifestsFromGlob');
            const promises = globs.map(pattern => Q.nfcall(glob, pattern));
            return Promise.all(promises)
                .then(results => _.uniq(_.flatten(results)))
                .then(results => {
                if (results.length > 0) {
                    trace.debug("Merging %s manifests from the following paths: ", results.length.toString());
                    results.forEach(path => trace.debug(path));
                }
                else {
                    throw new Error("No manifests found from the following glob patterns: \n" + this.settings.manifestGlobs.join("\n"));
                }
                return results;
            });
        }
        else {
            const manifests = this.settings.manifests;
            if (!manifests || manifests.length === 0) {
                return Q.reject("No manifests specified.");
            }
            this.settings.manifests = _.uniq(manifests).map(m => path.resolve(m));
            trace.debug("Merging %s manifest%s from the following paths: ", manifests.length.toString(), manifests.length === 1 ? "" : "s");
            manifests.forEach(path => trace.debug(path));
            return Q.resolve(this.settings.manifests);
        }
    }
    /**
     * Finds all manifests and merges them into two JS Objects: vsoManifest and vsixManifest
     * @return Q.Promise<SplitManifest> An object containing the two manifests
     */
    merge() {
        trace.debug('merger.merge');
        return this.gatherManifests().then(files => {
            let overridesProvided = false;
            let manifestPromises = [];
            files.forEach((file) => {
                manifestPromises.push(Q.nfcall(fs.readFile, file, "utf8").then((data) => {
                    let jsonData = data.replace(/^\uFEFF/, '');
                    try {
                        let result = JSON.parse(jsonData);
                        result.__origin = file; // save the origin in order to resolve relative paths later.
                        return result;
                    }
                    catch (err) {
                        trace.error("Error parsing the JSON in %s: ", file);
                        trace.debug(jsonData, null);
                        throw err;
                    }
                }));
            });
            // Add the overrides if necessary
            if (this.settings.overrides) {
                overridesProvided = true;
                manifestPromises.push(Q.resolve(this.settings.overrides));
            }
            return Promise.all(manifestPromises).then(partials => {
                // Determine the targets so we can construct the builders
                let targets = [];
                partials.forEach((partial) => {
                    if (_.isArray(partial["targets"])) {
                        targets = targets.concat(partial["targets"]);
                    }
                });
                this.extensionComposer = extension_composer_factory_1.ComposerFactory.GetComposer(this.settings, targets);
                this.manifestBuilders = this.extensionComposer.getBuilders();
                let updateVersionPromise = Promise.resolve(null);
                partials.forEach((partial, partialIndex) => {
                    // Rev the version if necessary
                    if (this.settings.revVersion) {
                        if (partial["version"] && partial.__origin) {
                            try {
                                const semver = version.SemanticVersion.parse(partial["version"]);
                                const newVersion = new version.SemanticVersion(semver.major, semver.minor, semver.patch + 1);
                                const newVersionString = newVersion.toString();
                                partial["version"] = newVersionString;
                                updateVersionPromise = qfs.readFile(partial.__origin, "utf8").then(versionPartial => {
                                    try {
                                        const newPartial = jsonInPlace(versionPartial).set("version", newVersionString);
                                        return qfs.writeFile(partial.__origin, newPartial);
                                    }
                                    catch (e) {
                                        trace.warn("Failed to lex partial as JSON to update the version. Skipping version rev...");
                                    }
                                });
                            }
                            catch (e) {
                                trace.warn("Could not parse %s as a semantic version (major.minor.patch). Skipping version rev...", partial["version"]);
                            }
                        }
                    }
                    // Transform asset paths to be relative to the root of all manifests, verify assets
                    if (_.isArray(partial["files"])) {
                        partial["files"].forEach((asset) => {
                            let keys = Object.keys(asset);
                            if (keys.indexOf("path") < 0) {
                                throw new Error("Files must have an absolute or relative (to the manifest) path.");
                            }
                            let absolutePath;
                            if (path.isAbsolute(asset.path)) {
                                absolutePath = asset.path;
                            }
                            else {
                                absolutePath = path.join(path.dirname(partial.__origin), asset.path);
                            }
                            asset.path = path.relative(this.settings.root, absolutePath);
                        });
                    }
                    // Transform icon paths as above
                    if (_.isObject(partial["icons"])) {
                        let icons = partial["icons"];
                        Object.keys(icons).forEach((iconKind) => {
                            let absolutePath = path.join(path.dirname(partial.__origin), icons[iconKind]);
                            icons[iconKind] = path.relative(this.settings.root, absolutePath);
                        });
                    }
                    // Expand any directories listed in the files array
                    if (_.isArray(partial["files"])) {
                        for (let i = partial["files"].length - 1; i >= 0; --i) {
                            let fileDecl = partial["files"][i];
                            let fsPath = path.join(this.settings.root, fileDecl.path);
                            if (fs.lstatSync(fsPath).isDirectory()) {
                                Array.prototype.splice.apply(partial["files"], [i, 1].concat(this.pathToFileDeclarations(fsPath, this.settings.root, fileDecl.addressable)));
                            }
                        }
                    }
                    // Process each key by each manifest builder.
                    Object.keys(partial).forEach((key) => {
                        let isOverridePartial = partials.length - 1 === partialIndex && overridesProvided;
                        if (partial[key] !== undefined && (partial[key] !== null || isOverridePartial)) {
                            // Notify each manifest builder of the key/value pair
                            this.manifestBuilders.forEach((builder) => {
                                builder.processKey(key, partial[key], isOverridePartial);
                            });
                        }
                    });
                });
                // Generate localization resources
                let locPrepper = new loc.LocPrep.LocKeyGenerator(this.manifestBuilders);
                let resources = locPrepper.generateLocalizationKeys();
                // Build up a master file list
                let packageFiles = {};
                this.manifestBuilders.forEach((builder) => {
                    _.assign(packageFiles, builder.files);
                });
                let components = { builders: this.manifestBuilders, resources: resources };
                // Finalize each builder
                return Promise.all([updateVersionPromise].concat(this.manifestBuilders.map(b => b.finalize(packageFiles, this.manifestBuilders)))).then(() => {
                    // Let the composer do validation
                    return this.extensionComposer.validate(components).then((validationResult) => {
                        if (validationResult.length === 0 || this.settings.bypassValidation) {
                            return components;
                        }
                        else {
                            throw new Error("There were errors with your extension. Address the following and re-run the tool.\n" + validationResult);
                        }
                    });
                });
            });
        });
    }
    /**
     * Recursively converts a given path to a flat list of FileDeclaration
     * @TODO: Async.
     */
    pathToFileDeclarations(fsPath, root, addressable) {
        let files = [];
        if (fs.lstatSync(fsPath).isDirectory()) {
            trace.debug("Path '%s` is a directory. Adding all contained files (recursive).", fsPath);
            fs.readdirSync(fsPath).forEach((dirChildPath) => {
                trace.debug("-- %s", dirChildPath);
                files = files.concat(this.pathToFileDeclarations(path.join(fsPath, dirChildPath), root, addressable));
            });
        }
        else {
            let relativePath = path.relative(root, fsPath);
            files.push({ path: relativePath, partName: "/" + relativePath, auto: true, addressable: addressable });
        }
        return files;
    }
}
exports.Merger = Merger;
