"use strict";
const _ = require("lodash");
const colors = require("colors");
const errHandler = require("../../../lib/errorhandler");
const fs = require("fs");
const GalleryInterfaces = require("vso-node-api/interfaces/GalleryInterfaces");
const Q = require("q");
const trace = require("../../../lib/trace");
const xml2js = require("xml2js");
const zip = require("jszip");
class GalleryBase {
    /**
     * Constructor
     * @param PublishSettings
     */
    constructor(settings, galleryClient, extInfo) {
        this.settings = settings;
        this.galleryClient = galleryClient;
        if (extInfo) {
            this.vsixInfoPromise = Q.resolve(extInfo);
        }
        // if (!settings.galleryUrl || !/^https?:\/\//.test(settings.galleryUrl)) {
        //     throw "Invalid or missing gallery URL.";
        // }
        // if (!settings.token || !/^[A-z0-9]{52}$/.test(settings.token)) {
        //     throw "Invalid or missing personal access token.";
        // }
    }
    getExtInfo() {
        if (!this.vsixInfoPromise) {
            this.vsixInfoPromise = GalleryBase.getExtInfo({
                extensionId: this.settings.extensionId,
                publisher: this.settings.publisher,
                vsixPath: this.settings.vsixPath });
        }
        return this.vsixInfoPromise;
    }
    static getExtInfo(info) {
        let promise;
        if (info.extensionId && info.publisher) {
            promise = Q.resolve({ id: info.extensionId, publisher: info.publisher, version: null });
        }
        else {
            promise = Q.Promise((resolve, reject, notify) => {
                fs.readFile(info.vsixPath, function (err, data) {
                    if (err)
                        reject(err);
                    trace.debug("Read vsix as zip... Size (bytes): %s", data.length.toString());
                    try {
                        resolve(new zip(data));
                    }
                    catch (err) {
                        reject(err);
                    }
                });
            }).then((zip) => {
                trace.debug("Files in the zip: %s", Object.keys(zip.files).join(", "));
                let vsixManifestFileNames = Object.keys(zip.files).filter(key => _.endsWith(key, "vsixmanifest"));
                if (vsixManifestFileNames.length > 0) {
                    return Q.nfcall(xml2js.parseString, zip.files[vsixManifestFileNames[0]].asText());
                }
                else {
                    throw "Could not locate vsix manifest!";
                }
            }).then((vsixManifestAsJson) => {
                let extensionId = info.extensionId || _.get(vsixManifestAsJson, "PackageManifest.Metadata[0].Identity[0].$.Id");
                let extensionPublisher = info.publisher || _.get(vsixManifestAsJson, "PackageManifest.Metadata[0].Identity[0].$.Publisher");
                let extensionVersion = _.get(vsixManifestAsJson, "PackageManifest.Metadata[0].Identity[0].$.Version");
                if (extensionId && extensionPublisher) {
                    return { id: extensionId, publisher: extensionPublisher, version: extensionVersion };
                }
                else {
                    throw "Could not locate both the extension id and publisher in vsix manfiest! Ensure your manifest includes both a namespace and a publisher property, or specify the necessary --publisher and/or --extension options.";
                }
            });
        }
        return promise;
    }
}
exports.GalleryBase = GalleryBase;
/**
 * Class that handles creating and deleting publishers
 */
class PublisherManager extends GalleryBase {
    /**
     * Constructor
     * @param PublishSettings
     */
    constructor(settings, galleryClient) {
        super(settings, galleryClient);
        this.settings = settings;
        this.galleryClient = galleryClient;
    }
    /**
     * Create a a publisher with the given name, displayName, and description
     * @param string Publisher's unique name
     * @param string Publisher's display name
     * @param string Publisher description
     * @return Q.Promise that is resolved when publisher is created
     */
    createPublisher(name, displayName, description) {
        return this.galleryClient.createPublisher({
            publisherName: name,
            displayName: displayName,
            longDescription: description,
            shortDescription: description
        }).catch(errHandler.httpErr);
    }
    /**
     * Delete the publisher with the given name
     * @param string Publisher's unique name
     * @return Q.promise that is resolved when publisher is deleted
     */
    deletePublisher(name) {
        return this.galleryClient.deletePublisher(name).catch(errHandler.httpErr);
    }
}
exports.PublisherManager = PublisherManager;
class SharingManager extends GalleryBase {
    shareWith(accounts) {
        return this.getExtInfo().then((extInfo) => {
            return Promise.all(accounts.map((account) => {
                trace.info("Sharing extension with %s.", account);
                return this.galleryClient.shareExtension(extInfo.publisher, extInfo.id, account).catch(errHandler.httpErr);
            }));
        });
    }
    unshareWith(accounts) {
        return this.getExtInfo().then((extInfo) => {
            return Promise.all(accounts.map((account) => {
                return this.galleryClient.unshareExtension(extInfo.publisher, extInfo.id, account).catch(errHandler.httpErr);
            }));
        });
    }
    unshareWithAll() {
        return this.getSharedWithAccounts().then((accounts) => {
            return this.unshareWith(accounts);
        });
    }
    getSharedWithAccounts() {
        return this.getExtensionInfo().then((ext) => {
            return ext.sharedWith.map(acct => acct.name);
        });
    }
    getExtensionInfo() {
        return this.getExtInfo().then((extInfo) => {
            return this.galleryClient.getExtension(extInfo.publisher, extInfo.id, null, GalleryInterfaces.ExtensionQueryFlags.IncludeVersions |
                GalleryInterfaces.ExtensionQueryFlags.IncludeFiles |
                GalleryInterfaces.ExtensionQueryFlags.IncludeCategoryAndTags |
                GalleryInterfaces.ExtensionQueryFlags.IncludeSharedAccounts).then((extension) => {
                return extension;
            }).catch(errHandler.httpErr);
        });
    }
}
exports.SharingManager = SharingManager;
class PackagePublisher extends GalleryBase {
    checkVsixPublished() {
        return this.getExtInfo().then((extInfo) => {
            return this.galleryClient.getExtension(extInfo.publisher, extInfo.id).then((ext) => {
                if (ext) {
                    extInfo.published = true;
                    return extInfo;
                }
                return extInfo;
            }).catch(() => extInfo);
        });
    }
    /**
     * Publish the VSIX extension given by vsixPath
     * @param string path to a VSIX extension to publish
     * @return Q.Promise that is resolved when publish is complete
     */
    publish() {
        let extPackage = {
            extensionManifest: fs.readFileSync(this.settings.vsixPath, "base64")
        };
        trace.debug("Publishing %s", this.settings.vsixPath);
        // Check if the app is already published. If so, call the update endpoint. Otherwise, create.
        trace.info("Checking if this extension is already published");
        return this.createOrUpdateExtension(extPackage).then((ext) => {
            trace.info("Waiting for server to validate extension package...");
            let versions = ext.versions;
            versions.sort((a, b) => {
                let aTime = a.lastUpdated.getTime();
                let bTime = b.lastUpdated.getTime();
                return aTime < bTime ? 1 : (aTime === bTime ? 0 : -1);
            });
            return this.waitForValidation(versions[0].version).then((result) => {
                if (result === PackagePublisher.validated) {
                    return ext;
                }
                else {
                    throw "Extension validation failed. Please address the following issues and retry publishing.\n" + result;
                }
            });
        });
    }
    createOrUpdateExtension(extPackage) {
        return this.checkVsixPublished().then((extInfo) => {
            let publishPromise;
            if (extInfo && extInfo.published) {
                trace.info("It is, %s the extension", colors.cyan("update").toString());
                publishPromise = this.galleryClient.updateExtension(extPackage, extInfo.publisher, extInfo.id).catch(errHandler.httpErr);
            }
            else {
                trace.info("It isn't, %s a new extension.", colors.cyan("create").toString());
                publishPromise = this.galleryClient.createExtension(extPackage).catch(errHandler.httpErr);
            }
            return publishPromise.then(() => {
                return this.galleryClient.getExtension(extInfo.publisher, extInfo.id, null, GalleryInterfaces.ExtensionQueryFlags.IncludeVersions);
            });
        });
    }
    waitForValidation(version, interval = PackagePublisher.validationInterval, retries = PackagePublisher.validationRetries) {
        if (retries === 0) {
            throw "Validation timed out. There may be a problem validating your extension. Please try again later.";
        }
        else if (retries === 25) {
            trace.info("This is taking longer than usual. Hold tight...");
        }
        trace.debug("Polling for validation (%s retries remaining).", retries.toString());
        // Compiler nonsense below. Sorry.
        return (Q.delay(this.getValidationStatus(version), interval)).then((status) => {
            trace.debug("--Retrieved validation status: %s", status);
            if (status === PackagePublisher.validationPending) {
                return this.waitForValidation(version, interval, retries - 1);
            }
            else {
                return Q.resolve(status); // otherwise TypeScript gets upset... I don't really know why.
            }
        });
    }
    getValidationStatus(version) {
        return this.getExtInfo().then((extInfo) => {
            return this.galleryClient.getExtension(extInfo.publisher, extInfo.id, extInfo.version, GalleryInterfaces.ExtensionQueryFlags.IncludeVersions).then((ext) => {
                if (!ext || ext.versions.length === 0) {
                    throw "Extension not published.";
                }
                let extVersion = ext.versions[0];
                if (version) {
                    extVersion = this.getVersionedExtension(ext, version);
                }
                // If there is a validationResultMessage, validation failed and this is the error
                // If the validated flag is missing and there is no validationResultMessage, validation is pending
                // If the validated flag is present and there is no validationResultMessage, the extension is validated.
                if (extVersion.validationResultMessage) {
                    return extVersion.validationResultMessage;
                }
                else if ((extVersion.flags & GalleryInterfaces.ExtensionVersionFlags.Validated) === 0) {
                    return PackagePublisher.validationPending;
                }
                else {
                    return PackagePublisher.validated;
                }
            });
        });
    }
    getVersionedExtension(extension, version) {
        let matches = extension.versions.filter(ev => ev.version === version);
        if (matches.length > 0) {
            return matches[0];
        }
        else {
            return null;
        }
    }
}
PackagePublisher.validationPending = "__validation_pending";
PackagePublisher.validated = "__validated";
PackagePublisher.validationInterval = 1000;
PackagePublisher.validationRetries = 50;
exports.PackagePublisher = PackagePublisher;
