"use strict";
const vsix_manifest_builder_1 = require("./vsix-manifest-builder");
const utils_1 = require("./utils");
const _ = require("lodash");
const fs = require("fs");
const mkdirp = require("mkdirp");
const os = require("os");
const path = require("path");
const Q = require("q");
const trace = require('../../../lib/trace');
const zip = require("jszip");
/**
 * Facilitates packaging the vsix and writing it to a file
 */
class VsixWriter {
    /**
     * constructor
     * @param any vsoManifest JS Object representing a vso manifest
     * @param any vsixManifest JS Object representing the XML for a vsix manifest
     */
    constructor(settings, components) {
        this.settings = settings;
        this.manifestBuilders = components.builders;
        this.resources = components.resources;
    }
    /**
     * If outPath is {auto}, generate an automatic file name.
     * Otherwise, try to determine if outPath is a directory (checking for a . in the filename)
     * If it is, generate an automatic filename in the given outpath
     * Otherwise, outPath doesn't change.
     */
    getOutputPath(outPath) {
        // Find the vsix manifest, if it exists
        let vsixBuilders = this.manifestBuilders.filter(b => b.getType() === vsix_manifest_builder_1.VsixManifestBuilder.manifestType);
        let autoName = "extension.vsix";
        if (vsixBuilders.length === 1) {
            let vsixManifest = vsixBuilders[0].getData();
            let pub = _.get(vsixManifest, "PackageManifest.Metadata[0].Identity[0].$.Publisher");
            let ns = _.get(vsixManifest, "PackageManifest.Metadata[0].Identity[0].$.Id");
            let version = _.get(vsixManifest, "PackageManifest.Metadata[0].Identity[0].$.Version");
            autoName = pub + "." + ns + "-" + version + ".vsix";
        }
        if (outPath === "{auto}") {
            return path.resolve(autoName);
        }
        else {
            let basename = path.basename(outPath);
            if (basename.indexOf(".") > 0) {
                return path.resolve(outPath);
            }
            else {
                return path.resolve(path.join(outPath, autoName));
            }
        }
    }
    static validatePartName(partName) {
        let segments = partName.split("/");
        if (segments.length === 1 && segments[0] === "[Content_Types].xml") {
            return true;
        }
        // matches most invalid segments.
        let re = /(%2f)|(%5c)|(^$)|(%[^0-9a-f])|(%.[^0-9a-f])|(\.$)|([^a-z0-9._~%!$&'()*+,;=:@-])/i;
        return segments.filter(segment => re.test(segment)).length === 0;
    }
    /**
     * Write a vsix package to the given file name
     */
    writeVsix() {
        let outputPath = this.getOutputPath(this.settings.outputPath);
        let vsix = new zip();
        let builderPromises = [];
        this.manifestBuilders.forEach((builder) => {
            // Avoid the error EMFILE: too many open files
            const addPackageFilesBatch = (paths, numBatch, batchSize, deferred) => {
                deferred = deferred || Q.defer();
                let readFilePromises = [];
                const start = numBatch * batchSize;
                const end = Math.min(paths.length, start + batchSize);
                for (let i = start; i < end; i++) {
                    const path = paths[i];
                    let itemName = utils_1.toZipItemName(builder.files[path].partName);
                    if (!VsixWriter.validatePartName(itemName)) {
                        let eol = require("os").EOL;
                        throw "Part Name '" + itemName + "' is invalid. Please check the following: " + eol + "1. No whitespace or any of these characters: #^[]<>?" + eol + "2. Cannot end with a period." + eol + "3. No percent-encoded / or \\ characters. Additionally, % must be followed by two hex characters.";
                    }
                    if (itemName.indexOf(" "))
                        if (!builder.files[path].content) {
                            let readFilePromise = Q.nfcall(fs.readFile, path).then((result) => {
                                vsix.file(itemName, result);
                            });
                            readFilePromises.push(readFilePromise);
                        }
                        else {
                            vsix.file(itemName, builder.files[path].content);
                            readFilePromises.push(Promise.resolve(null));
                        }
                }
                Promise.all(readFilePromises).then(function () {
                    if (end < paths.length) {
                        // Next batch
                        addPackageFilesBatch(paths, numBatch + 1, batchSize, deferred);
                    }
                    else {
                        deferred.resolve(null);
                    }
                }).catch(function (err) {
                    deferred.reject(err);
                });
                return deferred.promise;
            };
            // Add the package files in batches
            let builderPromise = addPackageFilesBatch(Object.keys(builder.files), 0, VsixWriter.VSIX_ADD_FILES_BATCH_SIZE).then(() => {
                // Add the manifest itself
                vsix.file(utils_1.toZipItemName(builder.getPath()), builder.getResult());
            });
            builderPromises.push(builderPromise);
        });
        return Promise.all(builderPromises).then(() => {
            return this.addResourceStrings(vsix);
        }).then(() => {
            trace.debug("Writing vsix to: %s", outputPath);
            return Q.nfcall(mkdirp, path.dirname(outputPath)).then(() => {
                let buffer = vsix.generate({
                    type: "nodebuffer",
                    compression: "DEFLATE"
                });
                return Q.nfcall(fs.writeFile, outputPath, buffer).then(() => outputPath);
            });
        });
    }
    /**
     * For each folder F under the localization folder (--loc-root),
     * look for a resources.resjson file within F. If it exists, split the
     * resources.resjson into one file per manifest. Add
     * each to the vsix archive as F/<manifest_loc_path> and F/Extension.vsixlangpack
     */
    addResourceStrings(vsix) {
        // Make sure locRoot is set, that it refers to a directory, and
        // iterate each subdirectory of that.
        if (!this.settings.locRoot) {
            return Promise.resolve(null);
        }
        let stringsPath = path.resolve(this.settings.locRoot);
        return Q.Promise((resolve, reject, notify) => {
            fs.exists(stringsPath, (exists) => {
                resolve(exists);
            });
        }).then((exists) => {
            if (exists) {
                return Q.nfcall(fs.lstat, stringsPath).then((stats) => {
                    if (stats.isDirectory()) {
                        return true;
                    }
                });
            }
            else {
                return Q.resolve(false);
            }
        }).then((stringsFolderExists) => {
            if (!stringsFolderExists) {
                return Promise.resolve(null);
            }
            return Q.nfcall(fs.readdir, stringsPath).then((files) => {
                let promises = [];
                files.forEach((languageTag) => {
                    var filePath = path.join(stringsPath, languageTag);
                    let promise = Q.nfcall(fs.lstat, filePath).then((fileStats) => {
                        if (fileStats.isDirectory()) {
                            let resourcePath = path.join(filePath, "resources.resjson");
                            return Q.Promise((resolve, reject, notify) => {
                                fs.exists(resourcePath, (exists) => {
                                    resolve(exists);
                                });
                            }).then((exists) => {
                                if (exists) {
                                }
                                else {
                                    return Promise.resolve(null);
                                }
                            });
                        }
                    });
                    promises.push(promise);
                });
                return Promise.all(promises);
            });
        });
    }
}
VsixWriter.VSIX_ADD_FILES_BATCH_SIZE = 20;
VsixWriter.VSO_MANIFEST_FILENAME = "extension.vsomanifest";
VsixWriter.VSIX_MANIFEST_FILENAME = "extension.vsixmanifest";
VsixWriter.CONTENT_TYPES_FILENAME = "[Content_Types].xml";
VsixWriter.DEFAULT_XML_BUILDER_SETTINGS = {
    indent: "    ",
    newline: os.EOL,
    pretty: true,
    xmldec: {
        encoding: "utf-8",
        standalone: null,
        version: "1.0"
    }
};
exports.VsixWriter = VsixWriter;
