"use strict";
const tfcommand_1 = require("../../lib/tfcommand");
const publish_1 = require("./_lib/publish");
const qfs_1 = require("../../lib/qfs");
const _ = require("lodash");
const args = require("../../lib/arguments");
const Q = require("q");
const trace = require("../../lib/trace");
function getCommand(args) {
    return new ExtensionBase(args);
}
exports.getCommand = getCommand;
class ManifestJsonArgument extends args.JsonArgument {
}
exports.ManifestJsonArgument = ManifestJsonArgument;
class ExtensionBase extends tfcommand_1.TfCommand {
    constructor(passedArgs, serverCommand = true) {
        super(passedArgs, serverCommand);
        this.description = "Commands to package, publish, and manage Extensions for Visual Studio Team Services.";
    }
    getHelpArgs() {
        return [];
    }
    setCommandArgs() {
        super.setCommandArgs();
        this.registerCommandArgument("extensionId", "Extension ID", "Use this as the extension ID instead of what is specified in the manifest.", args.StringArgument);
        this.registerCommandArgument("publisher", "Publisher name", "Use this as the publisher ID instead of what is specified in the manifest.", args.StringArgument);
        this.registerCommandArgument("serviceUrl", "Market URL", "URL to the VSS Marketplace.", args.StringArgument, "https://marketplace.visualstudio.com");
        this.registerCommandArgument("manifests", "Manifests", "List of individual manifest files (space separated).", args.ArrayArgument, "vss-extension.json");
        this.registerCommandArgument("manifestGlobs", "Manifest globs", "List of globs to find manifests (space separated).", args.ArrayArgument, null);
        this.registerCommandArgument("outputPath", "Output path", "Path to write the VSIX.", args.StringArgument, "{auto}");
        this.registerCommandArgument("override", "Overrides JSON", "JSON string which is merged into the manifests, overriding any values.", ManifestJsonArgument, "{}");
        this.registerCommandArgument("overridesFile", "Overrides JSON file", "Path to a JSON file with overrides. This partial manifest will always take precedence over any values in the manifests.", args.ReadableFilePathsArgument, null);
        this.registerCommandArgument("shareWith", "Share with", "List of VSTS Accounts with which to share the extension (space separated).", args.ArrayArgument, null);
        this.registerCommandArgument("unshareWith", "Un-share with", "List of VSTS Accounts with which to un-share the extension (space separated).", args.ArrayArgument, null);
        this.registerCommandArgument("vsix", "VSIX path", "Path to an existing VSIX (to publish or query for).", args.ReadableFilePathsArgument);
        this.registerCommandArgument("bypassValidation", "Bypass local validation", null, args.BooleanArgument, "false");
        this.registerCommandArgument("locRoot", "Localization root", "Root of localization hierarchy (see README for more info).", args.ExistingDirectoriesArgument, null);
        this.registerCommandArgument("displayName", "Display name", null, args.StringArgument);
        this.registerCommandArgument("description", "Description", "Description of the Publisher.", args.StringArgument);
        this.registerCommandArgument("revVersion", "Rev version", "Rev the patch-version of the extension and save the result.", args.BooleanArgument, "false");
    }
    getMergeSettings() {
        return Promise.all([
            this.commandArgs.root.val(),
            this.commandArgs.manifests.val(),
            this.commandArgs.manifestGlobs.val(),
            this.commandArgs.override.val(),
            this.commandArgs.overridesFile.val(),
            this.commandArgs.revVersion.val(),
            this.commandArgs.bypassValidation.val(),
            this.commandArgs.publisher.val(true),
            this.commandArgs.extensionId.val(true)
        ]).then((values) => {
            const [root, manifests, manifestGlob, override, overridesFile, revVersion, bypassValidation, publisher, extensionId] = values;
            if (publisher) {
                _.set(override, "publisher", publisher);
            }
            if (extensionId) {
                _.set(override, "extensionid", extensionId);
            }
            let overrideFileContent = Q.resolve("");
            if (overridesFile && overridesFile.length > 0) {
                overrideFileContent = qfs_1.readFile(overridesFile[0], "utf-8");
            }
            return overrideFileContent.then((contentStr) => {
                let content = contentStr;
                if (content === "") {
                    content = "{}";
                    if (overridesFile && overridesFile.length > 0) {
                        trace.warn("Overrides file was empty. No overrides will be imported from " + overridesFile[0]);
                    }
                }
                let mergedOverrides = {};
                let contentJSON = "";
                try {
                    contentJSON = JSON.parse(content);
                }
                catch (e) {
                    throw "Could not parse contents of " + overridesFile[0] + " as JSON. \n" + e;
                }
                _.merge(mergedOverrides, contentJSON, override);
                return {
                    root: root[0],
                    manifests: manifests,
                    manifestGlobs: manifestGlob,
                    overrides: mergedOverrides,
                    bypassValidation: bypassValidation,
                    revVersion: revVersion
                };
            });
        });
    }
    getPackageSettings() {
        return Promise.all([
            this.commandArgs.outputPath.val(),
            this.commandArgs.locRoot.val()
        ]).then((values) => {
            const [outputPath, locRoot] = values;
            return {
                outputPath: outputPath,
                locRoot: locRoot && locRoot[0]
            };
        });
    }
    identifyExtension() {
        return this.commandArgs.vsix.val(true).then((result) => {
            let vsixPath = _.isArray(result) ? result[0] : null;
            let infoPromise;
            if (!vsixPath) {
                infoPromise = Promise.all([this.commandArgs.publisher.val(), this.commandArgs.extensionId.val()]).then((values) => {
                    const [publisher, extensionId] = values;
                    return publish_1.GalleryBase.getExtInfo({ extensionId: extensionId, publisher: publisher });
                });
            }
            else {
                infoPromise = Promise.all([
                    this.commandArgs.publisher.val(true),
                    this.commandArgs.extensionId.val(true)
                ]).then((values) => {
                    const [publisher, extensionId] = values;
                    return publish_1.GalleryBase.getExtInfo({ vsixPath: vsixPath, publisher: publisher, extensionId: extensionId });
                });
            }
            return infoPromise;
        });
    }
    getPublishSettings() {
        return Promise.all([
            this.commandArgs.serviceUrl.val(),
            this.commandArgs.vsix.val(true),
            this.commandArgs.publisher.val(true),
            this.commandArgs.extensionId.val(true),
            this.commandArgs.shareWith.val()
        ]).then((values) => {
            const [marketUrl, vsix, publisher, extensionId, shareWith] = values;
            let vsixPath = _.isArray(vsix) ? vsix[0] : null;
            return {
                galleryUrl: marketUrl,
                vsixPath: vsixPath,
                publisher: publisher,
                extensionId: extensionId,
                shareWith: shareWith
            };
        });
    }
    exec(cmd) {
        return this.getHelp(cmd);
    }
}
exports.ExtensionBase = ExtensionBase;
