"use strict";
const args = require("../../lib/arguments");
const colors = require("colors");
const extBase = require("./default");
const extInfo = require("./_lib/extensioninfo");
const trace = require("../../lib/trace");
const SPS_INSTANCE_TYPE = "951917AC-A960-4999-8464-E3F0AA25B381";
function getCommand(args) {
    return new ExtensionInstall(args);
}
exports.getCommand = getCommand;
class AccountInstallReport {
    constructor(itemId, accountName, accountId, installed = false, reason) {
        this.itemId = itemId;
        this.accountName = accountName;
        this.accountId = accountId;
        this.installed = installed;
        this.reason = reason;
    }
    setError(reason) {
        this.installed = false;
        this.reason = reason;
    }
    setInstalled(reason) {
        this.installed = true;
        this.reason = reason;
    }
}
exports.AccountInstallReport = AccountInstallReport;
class ExtensionInstall extends extBase.ExtensionBase {
    constructor(passedArgs) {
        super(passedArgs);
        this.description = "Install a Visual Studio Services Extension to a list of VSTS Accounts.";
    }
    setCommandArgs() {
        super.setCommandArgs();
        this.registerCommandArgument("accounts", "Installation target accounts", "List of accounts where to install the extension.", args.ArrayArgument);
    }
    getHelpArgs() {
        return ["publisher", "extensionId", "vsix", "accounts"];
    }
    exec() {
        // Read extension info from arguments
        const result = { accounts: {}, extension: null };
        return this._getExtensionInfo()
            .then(extInfo => {
            const itemId = `${extInfo.publisher}.${extInfo.id}`;
            const galleryApi = this.webApi.getGalleryApi(this.webApi.serverUrl);
            result.extension = itemId;
            // Read accounts from arguments and resolve them to get its accountIds
            return this.commandArgs.accounts.val().then(accounts => {
                // Install extension in each account
                const installations = [...accounts].map((account) => {
                    const emsApi = this.webApi.getExtensionManagementApi(this.getEmsAccountUrl(this.webApi.serverUrl, account));
                    return emsApi.installExtensionByName(extInfo.publisher, extInfo.id).then(installation => [account, installation]);
                });
                return Promise.all(installations);
            }).then(installations => {
                installations.forEach(installation => {
                    const account = installation[0];
                    const installedExtension = installation[1];
                    const installationResult = { installed: true, issues: null };
                    if (installedExtension.installState.installationIssues && installedExtension.installState.installationIssues.length > 0) {
                        installationResult.installed = false;
                        installationResult.issues = `The following issues were encountered installing to ${account}: 
${installedExtension.installState.installationIssues.map(i => " - " + i).join("\n")}`;
                    }
                    result.accounts[account] = installationResult;
                });
                return result;
            });
        });
    }
    getEmsAccountUrl(marketplaceUrl, accountName) {
        if (marketplaceUrl.toLocaleLowerCase().indexOf("marketplace.visualstudio.com") >= 0) {
            return `https://${accountName}.extmgmt.visualstudio.com`;
        }
        if (marketplaceUrl.toLocaleLowerCase().indexOf("me.tfsallin.net") >= 0) {
            return marketplaceUrl.toLocaleLowerCase().indexOf("https://") === 0 ?
                `https://${accountName}.me.tfsallin.net:8781` :
                `http://${accountName}.me.tfsallin.net:8780`;
        }
        return marketplaceUrl;
    }
    friendlyOutput(data) {
        trace.success("\n=== Completed operation: install extension ===");
        Object.keys(data.accounts).forEach(a => {
            trace.info(`- ${a}: ${data.accounts[a].installed ? colors.green("success") : colors.red(data.accounts[a].issues)}`);
        });
    }
    _getExtensionInfo() {
        return this.commandArgs.vsix.val(true).then((vsixPath) => {
            let extInfoPromise;
            if (vsixPath !== null) {
                extInfoPromise = extInfo.getExtInfo(vsixPath[0], null, null);
            }
            else {
                extInfoPromise = Promise.all([
                    this.commandArgs.publisher.val(),
                    this.commandArgs.extensionId.val()]).then((values) => {
                    const [publisher, extension] = values;
                    return extInfo.getExtInfo(null, extension, publisher);
                });
            }
            return extInfoPromise;
        });
    }
}
exports.ExtensionInstall = ExtensionInstall;
