"use strict";
const diskcache_1 = require("../lib/diskcache");
const credstore_1 = require("../lib/credstore");
const common_1 = require("../lib/common");
const connection_1 = require("../lib/connection");
const WebApi_1 = require("vso-node-api/WebApi");
const os_1 = require("os");
const _ = require("lodash");
const args = require("./arguments");
const colors_1 = require("colors");
const common = require("./common");
const copypaste = require("copy-paste");
const loader = require("../lib/loader");
const path = require("path");
const Q = require("q");
const qfs = require("./qfs");
const trace = require("./trace");
const version = require("./version");
class TfCommand {
    /**
     * @param serverCommand True to initialize the WebApi object during init phase.
     */
    constructor(passedArgs, serverCommand = true) {
        this.passedArgs = passedArgs;
        this.serverCommand = serverCommand;
        this.commandArgs = {};
        this.description = "A suite of command line tools to interact with Visual Studio Team Services.";
        this.setCommandArgs();
    }
    /**
     * Returns a promise that is resolved when this command is initialized and
     * ready to be executed.
     */
    ensureInitialized() {
        return this.initialized || this.initialize();
    }
    initialize() {
        this.initialized = this.commandArgs.help.val().then((needHelp) => {
            if (needHelp) {
                return this.run.bind(this, this.getHelp.bind(this));
            }
            else {
                // Set the fiddler proxy
                return this.commandArgs.fiddler.val().then((useProxy) => {
                    if (useProxy) {
                        process.env.HTTP_PROXY = "http://127.0.0.1:8888";
                    }
                }).then(() => {
                    // Set custom proxy 
                    return this.commandArgs.proxy.val(true).then((proxy) => {
                        if (proxy) {
                            process.env.HTTP_PROXY = proxy;
                        }
                    });
                }).then(() => {
                    // Set the no-prompt flag 
                    return this.commandArgs.noPrompt.val(true).then((noPrompt) => {
                        common.NO_PROMPT = noPrompt;
                    });
                }).then(() => {
                    // Set the cached service url
                    return this.commandArgs.serviceUrl.val(true).then((serviceUrl) => {
                        if (!serviceUrl && !process.env["TFX_BYPASS_CACHE"] && common.EXEC_PATH.join("") !== "login") {
                            let diskCache = new diskcache_1.DiskCache("tfx");
                            return diskCache.itemExists("cache", "connection").then((isConnection) => {
                                let connectionUrlPromise;
                                if (!isConnection) {
                                    connectionUrlPromise = Promise.resolve(null);
                                }
                                else {
                                    connectionUrlPromise = diskCache.getItem("cache", "connection");
                                }
                                return connectionUrlPromise.then((url) => {
                                    if (url) {
                                        this.commandArgs.serviceUrl.setValue(url);
                                    }
                                });
                            });
                        }
                        else {
                            return Promise.resolve(null);
                        }
                    });
                }).then(() => {
                    let apiPromise = Promise.resolve(null);
                    if (this.serverCommand) {
                        apiPromise = this.getWebApi().then(_ => { });
                    }
                    return apiPromise.then(() => {
                        return this.run.bind(this, this.exec.bind(this));
                    });
                });
            }
        });
        return this.initialized;
    }
    getGroupedArgs() {
        if (!this.groupedArgs) {
            let group = {};
            let currentArg = null;
            this.passedArgs.forEach((arg) => {
                if (_.startsWith(arg, "--")) {
                    currentArg = _.camelCase(arg.substr(2));
                    group[currentArg] = [];
                    return;
                }
                if (currentArg) {
                    group[currentArg].push(arg);
                }
            });
            this.groupedArgs = group;
        }
        return this.groupedArgs;
    }
    registerCommandArgument(name, friendlyName, description, ctor, defaultValue) {
        let groupedArgs = this.getGroupedArgs();
        if (groupedArgs[name]) {
            this.commandArgs[name] = new ctor(name, friendlyName, description, groupedArgs[name]);
        }
        else {
            this.commandArgs[name] = new ctor(name, friendlyName, description, defaultValue, true);
        }
    }
    /**
     * Register arguments that may be used with this command.
     */
    setCommandArgs() {
        this.registerCommandArgument("project", "Project name", null, args.StringArgument);
        this.registerCommandArgument("root", "Root directory", null, args.ExistingDirectoriesArgument, ".");
        this.registerCommandArgument("authType", "Authentication Method", "Method of authentication ('pat' or 'basic').", args.StringArgument, "pat");
        this.registerCommandArgument("serviceUrl", "Service URL", "URL to the service you will connect to, e.g. https://youraccount.visualstudio.com/DefaultCollection.", args.StringArgument);
        this.registerCommandArgument("password", "Password", "Password to use for basic authentication.", args.SilentStringArgument);
        this.registerCommandArgument("token", "Personal access token", null, args.SilentStringArgument);
        this.registerCommandArgument("save", "Save settings", "Save arguments for the next time a command in this command group is run.", args.BooleanArgument, "false");
        this.registerCommandArgument("username", "Username", "Username to use for basic authentication.", args.StringArgument);
        this.registerCommandArgument("output", "Output destination", "Method to use for output. Options: friendly, json, clipboard.", args.StringArgument, "friendly");
        this.registerCommandArgument("json", "Output as JSON", "Alias for --output json.", args.BooleanArgument, "false");
        this.registerCommandArgument("fiddler", "Use Fiddler proxy", "Set up the fiddler proxy for HTTP requests (for debugging purposes).", args.BooleanArgument, "false");
        this.registerCommandArgument("proxy", "Proxy server", "Use the specified proxy server for HTTP traffic.", args.StringArgument, null);
        this.registerCommandArgument("help", "Help", "Get help for any command.", args.BooleanArgument, "false");
        this.registerCommandArgument("noPrompt", "No Prompt", "Do not prompt the user for input (instead, raise an error).", args.BooleanArgument, "false");
    }
    /**
     * Return a list of registered arguments that should be displayed when help is emitted.
     */
    getHelpArgs() {
        return [];
    }
    /**
     * Get a BasicCredentialHandler based on the command arguments:
     * If username & password are passed in, use those.
     * If token is passed in, use that.
     * Else, check the authType - if it is "pat", prompt for a token
     * If it is "basic", prompt for username and password.
     */
    getCredentials(serviceUrl, useCredStore = true) {
        return Promise.all([
            this.commandArgs.authType.val(),
            this.commandArgs.token.val(true),
            this.commandArgs.username.val(true),
            this.commandArgs.password.val(true)
        ]).then((values) => {
            const [authType, token, username, password] = values;
            if (username && password) {
                return WebApi_1.getBasicHandler(username, password);
            }
            else {
                if (token) {
                    return WebApi_1.getBasicHandler("OAuth", token);
                }
                else {
                    let getCredentialPromise;
                    if (useCredStore) {
                        getCredentialPromise = credstore_1.getCredentialStore("tfx").getCredential(serviceUrl, "allusers");
                    }
                    else {
                        getCredentialPromise = Q.reject();
                    }
                    return getCredentialPromise.then((credString) => {
                        if (credString.length <= 6) {
                            throw "Could not get credentials from credential store.";
                        }
                        if (credString.substr(0, 3) === "pat") {
                            return WebApi_1.getBasicHandler("OAuth", credString.substr(4));
                        }
                        else if (credString.substr(0, 5) === "basic") {
                            let rest = credString.substr(6);
                            let unpwDividerIndex = rest.indexOf(":");
                            let username = rest.substr(0, unpwDividerIndex);
                            let password = rest.substr(unpwDividerIndex + 1);
                            if (username && password) {
                                return WebApi_1.getBasicHandler(username, password);
                            }
                            else {
                                throw "Could not get credentials from credential store.";
                            }
                        }
                    }).catch(() => {
                        if (authType.toLowerCase() === "pat") {
                            return this.commandArgs.token.val().then((token) => {
                                return WebApi_1.getBasicHandler("OAuth", token);
                            });
                        }
                        else if (authType.toLowerCase() === "basic") {
                            return this.commandArgs.username.val().then((username) => {
                                return this.commandArgs.password.val().then((password) => {
                                    return WebApi_1.getBasicHandler(username, password);
                                });
                            });
                        }
                        else {
                            throw new Error("Unsupported auth type. Currently, 'pat' and 'basic' auth are supported.");
                        }
                    });
                }
            }
        });
    }
    getWebApi() {
        return this.commandArgs.serviceUrl.val().then((url) => {
            return this.getCredentials(url).then((handler) => {
                this.connection = new connection_1.TfsConnection(url);
                this.webApi = new WebApi_1.WebApi(url, handler);
                return this.webApi;
            });
        });
    }
    run(main, cmd) {
        return main(cmd).then((result) => {
            return this.output(result).then(() => {
                return this.dispose();
            });
        });
    }
    /**
     * Should be called after exec. In here we will write settings to fs if necessary.
     */
    dispose() {
        let newToCache = {};
        return this.commandArgs.save.val().then((shouldSave) => {
            if (shouldSave) {
                let cacheKey = path.resolve().replace("/\.\[\]/g", "-") + "." +
                    common.EXEC_PATH.slice(0, common.EXEC_PATH.length - 1).join("/");
                let getValuePromises = [];
                Object.keys(this.commandArgs).forEach((arg) => {
                    let argObj = this.commandArgs[arg];
                    if (!argObj.hasDefaultValue) {
                        let pr = argObj.val().then((value) => {
                            // don"t cache these 5 options.
                            if (["username", "password", "save", "token", "help"].indexOf(arg) < 0) {
                                _.set(newToCache, cacheKey + "." + arg, value);
                            }
                        });
                        getValuePromises.push(pr);
                    }
                });
                return Promise.all(getValuePromises).then(() => {
                    return args.getOptionsCache().then((existingCache) => {
                        // custom shallow-ish merge of cache properties.
                        let newInThisCommand = _.get(newToCache, cacheKey);
                        if (!_.get(existingCache, cacheKey)) {
                            _.set(existingCache, cacheKey, {});
                        }
                        if (newInThisCommand) {
                            Object.keys(newInThisCommand).forEach((key) => {
                                _.set(existingCache, cacheKey + "." + key, newInThisCommand[key]);
                            });
                            new diskcache_1.DiskCache("tfx").setItem("cache", "command-options", JSON.stringify(existingCache, null, 4).replace(/\n/g, os_1.EOL));
                        }
                    });
                });
            }
            else {
                return Promise.resolve(null);
            }
        });
    }
    /**
     * Gets help (as a string) for the given command
     */
    getHelp(cmd) {
        this.commandArgs.output.setValue("help");
        let result = os_1.EOL;
        result += ["                        fTfs         ",
            "                      fSSSSSSSs      ",
            "                    fSSSSSSSSSS      ",
            "     TSSf         fSSSSSSSSSSSS      ",
            "     SSSSSF     fSSSSSSST SSSSS      ",
            "     SSfSSSSSsfSSSSSSSt   SSSSS      ",
            "     SS  tSSSSSSSSSs      SSSSS      ",
            "     SS   fSSSSSSST       SSSSS      ",
            "     SS fSSSSSFSSSSSSf    SSSSS      ",
            "     SSSSSST    FSSSSSSFt SSSSS      ",
            "     SSSSt        FSSSSSSSSSSSS      ",
            "                    FSSSSSSSSSS      ",
            "                       FSSSSSSs      ",
            "                        FSFs    (TM) "].
            map(l => colors_1.magenta(l)).join(os_1.EOL) + os_1.EOL + os_1.EOL;
        let continuedHierarchy = cmd.commandHierarchy;
        cmd.execPath.forEach((segment) => {
            continuedHierarchy = continuedHierarchy[segment];
        });
        if (continuedHierarchy === null) {
            // Need help with a particular command
            let singleArgData = (argName, maxArgLen) => {
                let argKebab = _.kebabCase(argName);
                let argObj = this.commandArgs[argName];
                return "  --" +
                    argKebab + "  " +
                    common_1.repeatStr(" ", maxArgLen - argKebab.length) +
                    colors_1.gray((argObj.description || (argObj.friendlyName + "."))) + os_1.EOL;
            };
            let commandName = cmd.execPath[cmd.execPath.length - 1];
            result += colors_1.cyan("Syntax: ") + os_1.EOL +
                colors_1.cyan("tfx ") + colors_1.yellow(cmd.execPath.join(" ")) +
                colors_1.green(" --arg1 arg1val1 arg1val2[...]") +
                colors_1.gray(" --arg2 arg2val1 arg2val2[...]") + os_1.EOL + os_1.EOL;
            return loader.load(cmd.execPath, []).then((tfCommand) => {
                result += colors_1.cyan("Command: ") + commandName + os_1.EOL;
                result += tfCommand.description + os_1.EOL + os_1.EOL;
                result += colors_1.cyan("Arguments: ") + os_1.EOL;
                let uniqueArgs = this.getHelpArgs();
                uniqueArgs = _.uniq(uniqueArgs);
                let maxArgLen = uniqueArgs.map(a => _.kebabCase(a)).reduce((a, b) => Math.max(a, b.length), 0);
                if (uniqueArgs.length === 0) {
                    result += "[No arguments for this command]" + os_1.EOL;
                }
                uniqueArgs.forEach((arg) => {
                    result += singleArgData(arg, maxArgLen);
                });
                if (this.serverCommand) {
                    result += os_1.EOL + colors_1.cyan("Global server command arguments:") + os_1.EOL;
                    ["authType", "username", "password", "token", "serviceUrl", "fiddler", "proxy"].forEach((arg) => {
                        result += singleArgData(arg, 11);
                    });
                }
                result += os_1.EOL + colors_1.cyan("Global arguments:") + os_1.EOL;
                ["help", "save", "noPrompt", "output", "json"].forEach((arg) => {
                    result += singleArgData(arg, 9);
                });
                result += os_1.EOL + colors_1.gray("To see more commands, type " + colors_1.reset("tfx " + cmd.execPath.slice(0, cmd.execPath.length - 1).join(" ") + " --help"));
            }).then(() => {
                return result;
            });
        }
        else {
            // Need help with a suite of commands
            // There is a weird coloring bug when colors are nested, so we don"t do that.
            result += colors_1.cyan("Available ") +
                "commands" +
                colors_1.cyan(" and ") +
                colors_1.yellow("command groups") +
                colors_1.cyan(" in " + ["tfx"].concat(cmd.execPath).join(" / ") + ":") + os_1.EOL;
            let commandDescriptionPromises = [];
            Object.keys(continuedHierarchy).forEach((command) => {
                if (command === "default") {
                    return;
                }
                let pr = loader.load(cmd.execPath.concat([command]), []).then((tfCommand) => {
                    let coloredCommand = command;
                    if (continuedHierarchy[command] !== null) {
                        coloredCommand = colors_1.yellow(command);
                    }
                    result += " - " + coloredCommand + colors_1.gray(": " + tfCommand.description) + os_1.EOL;
                });
                commandDescriptionPromises.push(pr);
            });
            return Promise.all(commandDescriptionPromises).then(() => {
                result += os_1.EOL + os_1.EOL + colors_1.gray("For help with an individual command, type ") + colors_1.reset("tfx " + cmd.execPath.join(" ") + " <command> --help") + os_1.EOL;
            }).then(() => {
                return result;
            });
        }
    }
    /**
     * Display a copyright banner.
     */
    showBanner() {
        return this.commandArgs.json.val(true).then((useJson) => {
            if (useJson) {
                this.commandArgs.output.setValue("json");
            }
        }).then(() => {
            return this.commandArgs.output.val(true).then((outputType) => {
                return version.getTfxVersion().then((semVer) => {
                    trace.outputType = outputType;
                    if (outputType === "friendly") {
                        trace.info(colors_1.gray("TFS Cross Platform Command Line Interface v" + semVer.toString()));
                        trace.info(colors_1.gray("Copyright Microsoft Corporation"));
                    }
                });
            });
        });
    }
    /**
     * Takes data and pipes it to the appropriate output mechanism
     */
    output(data) {
        return this.commandArgs.output.val().then((outputDestination) => {
            switch (outputDestination.toLowerCase()) {
                case "friendly":
                    this.friendlyOutput(data);
                    break;
                case "json":
                    this.jsonOutput(data);
                    break;
                case "help":
                    this.friendlyOutputConstant(data);
                    break;
                case "clip":
                case "clipboard":
                    let clipboardText = this.getClipboardOutput(data);
                    return Q.nfcall(copypaste.copy, clipboardText);
                default:
                    return qfs.canWriteTo(path.resolve(outputDestination)).then((canWrite) => {
                        if (canWrite) {
                            let fileContents = this.getFileOutput(data);
                            return qfs.writeFile(outputDestination, fileContents);
                        }
                        else {
                            throw new Error("Cannot write output to " + outputDestination);
                        }
                    });
            }
            return Promise.resolve(null);
        });
    }
    /**
     * Given the output object, gets the string that is copied to the clipboard when
     * clipboard output is requested.
     */
    getClipboardOutput(data) {
        return this.getOutputString(data);
    }
    /**
     * Given the output object, gets the string that is written to a destination
     * file when a file name is given as the output destination
     */
    getFileOutput(data) {
        return this.getOutputString(data);
    }
    getOutputString(data) {
        let outputString = "";
        try {
            outputString = JSON.stringify(data, null, 4);
        }
        catch (e) {
            if (data && data.toString) {
                outputString = data.toString();
            }
            else {
                outputString = data + "";
            }
        }
        return outputString;
    }
    /**
     * Gets a nicely formatted output string for friendly output
     */
    friendlyOutput(data) {
        this.friendlyOutputConstant(data);
    }
    friendlyOutputConstant(data) {
        if (typeof data === "string") {
            console.log(data);
        }
        else {
            try {
                console.log(JSON.stringify(data, null, 4));
            }
            catch (e) {
                console.log(data + "");
            }
        }
    }
    /**
     * Gets a string of valid JSON when JSON output is requested.
     * Probably no need to override this one.
     */
    jsonOutput(data) {
        try {
            console.log(colors_1.reset(JSON.stringify(data, null, 4)));
        }
        catch (e) {
            throw new Error("Could not stringify JSON output.");
        }
    }
}
exports.TfCommand = TfCommand;
